import time
import unittest
import PIL.Image
import torch
import torchvision
from irisml.tasks.benchmark_model import Task


class FakeDataset:
    def __init__(self, num_samples):
        self._num_samples = num_samples

    def __len__(self):
        return self._num_samples

    def __getitem__(self, index):
        return PIL.Image.new('RGB', (224, 224)), torch.tensor(1)


class TestBenchmarkModel(unittest.TestCase):
    def test_simple_model(self):
        class FakeModel(torch.nn.Module):
            def training_step(self, inputs, targets):
                time.sleep(0.01)
                return {'loss': torch.tensor(1.0, requires_grad=True)}

            def prediction_step(self, inputs):
                time.sleep(0.01)
                return

        model = FakeModel()
        dataset = FakeDataset(100)
        transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(1, 'cpu')).execute(Task.Inputs(model, dataset, transform))

        self.assertGreater(outputs.forward_time_per_iteration, 0.01)
        self.assertGreater(outputs.forward_backward_time_per_iteration, 0.01)
        self.assertGreater(outputs.prediction_time_per_iteration, 0.01)
        self.assertEqual(outputs.max_cuda_memory_in_mb, 0)  # On CPU

    def test_model_without_step_method(self):
        class FakeModel(torch.nn.Module):
            @property
            def criterion(self):
                return torch.nn.CrossEntropyLoss()

            def forward(self, x):
                time.sleep(0.01)
                return torch.zeros((x.shape[0], 8), requires_grad=True)

        model = FakeModel()
        dataset = FakeDataset(100)
        transform = torchvision.transforms.ToTensor()
        outputs = Task(Task.Config(1, 'cpu')).execute(Task.Inputs(model, dataset, transform))
        self.assertGreater(outputs.forward_time_per_iteration, 0.01)



import copy
import dataclasses
import functools
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Make a wrapper model to extract a feature vector from a vision model.

    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        output_module_name: typing.Optional[str] = None

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None

    def execute(self, inputs):
        model = FeatureExtractor(copy.deepcopy(inputs.model), self.config.output_module_name)
        return self.Outputs(model)

    def dry_run(self, inputs):
        return self.execute(inputs)


class FeatureExtractor(torch.nn.Module):
    def __init__(self, model, output_module_name=None):
        """Extract a feature vector from a model.

        Notes:
            - This class doesn't support multi-thread inference.
            - If the output module was used twice, the first feature vector will be returned.

        Args:
            model (torch.nn.Module): A model. Must not be re-used in other place.
            output_module_name: The name of the module to extract feature vectors.
        """
        super().__init__()
        self._model = model
        self._output_module_name = output_module_name

        if output_module_name:
            module = self._find_module(self._model, output_module_name)
            # Instead of having a plugin to remove the hook, we forget about it. The model shouldn't be reused.
            module.register_forward_hook(self._save_outputs)

    def prediction_step(self, inputs):
        self._features = None
        outputs = self._model(inputs)
        outputs = self._features if self._output_module_name else outputs
        self._features = None
        return outputs

    def _save_outputs(self, module, inputs, outputs):
        if self._features is None:
            self._features = copy.deepcopy(outputs.detach())

    @staticmethod
    def _find_module(model, name):
        names = name.split('.')
        return functools.reduce(getattr, names, model)



import dataclasses
import logging
import statistics
import time
import typing
import torch
import irisml.core
from irisml.tasks.train.build_dataloader import build_dataloader

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Benchmark a given model using a given dataset."""
    VERSION = '0.1.0'
    CACHE_ENABLED = False

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        dataset: torch.utils.data.Dataset
        transform: typing.Callable

    @dataclasses.dataclass
    class Config:
        batch_size: int
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        num_iterations: int = 10

    @dataclasses.dataclass
    class Outputs:
        forward_backward_time_per_iteration: float = 0.0
        forward_time_per_iteration: float = 0.0
        backward_time_per_iteration: float = 0.0
        prediction_time_per_iteration: float = 0.0
        max_cuda_memory_in_mb: float = 0.0

    def execute(self, inputs):
        device = self._get_device()
        dataloader = build_dataloader(inputs.dataset, inputs.transform, batch_size=self.config.batch_size, shuffle=False, drop_last=False)
        forward_time_all = []
        backward_time_all = []
        prediction_time_all = []

        model = inputs.model.to(device)
        model.train()

        if device.type == 'cuda':
            torch.cuda.reset_peak_memory_stats(device)
            cuda_memory_at_start = torch.cuda.memory_allocated(device)

        # Measure training forward and backward pass
        for i, batch in enumerate(dataloader):
            inputs, targets = batch
            inputs = self._to_device(inputs, device)
            targets = self._to_device(targets, device)
            forward_time, backward_time = self._training_step(model, inputs, targets)
            forward_time_all.append(forward_time)
            backward_time_all.append(backward_time)
            if i >= self.config.num_iterations:
                break

        model.eval()
        # Measure prediction time.
        for i, batch in enumerate(dataloader):
            inputs, targets = batch
            inputs = self._to_device(inputs, device)
            prediction_time_all.append(self._prediction_step(model, inputs))
            if i >= self.config.num_iterations:
                break

        if len(forward_time_all) < self.config.num_iterations:
            logger.info(f"The dataset is smaller than expected. The actual number of iteration is {len(forward_time_all)}")

        forward_time_per_iteration = statistics.fmean(forward_time_all)
        backward_time_per_iteration = statistics.fmean(backward_time_all)
        forward_backward_time_per_iteration = forward_time_per_iteration + backward_time_per_iteration
        prediction_time_per_iteration = statistics.fmean(prediction_time_all)
        max_cuda_memory_in_mb = ((torch.cuda.max_memory_allocated(device) - cuda_memory_at_start) / 2 ** 20) if device.type == 'cuda' else 0

        logger.info(f"{forward_time_per_iteration=}, {backward_time_per_iteration=}, {forward_backward_time_per_iteration=}, {prediction_time_per_iteration=}")
        if max_cuda_memory_in_mb > 0:
            logger.info(f"{max_cuda_memory_in_mb=}")

        return self.Outputs(forward_backward_time_per_iteration, forward_time_per_iteration, backward_time_per_iteration, prediction_time_per_iteration, max_cuda_memory_in_mb)

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)

    def _to_device(self, data, device):
        if isinstance(data, list):
            return [self._to_device(d) for d in data]
        elif isinstance(data, tuple):
            return tuple(self._to_device(d) for d in data)
        elif hasattr(data, 'to'):
            return data.to(device, non_blocking=True)
        return data

    @staticmethod
    def _synchronize(device):
        if device.type == 'cuda':
            torch.cuda.synchronize()

    def _training_step(self, model, inputs, targets):
        self._synchronize(inputs.device)

        start = time.time()
        if hasattr(model, 'training_step'):
            loss = model.training_step(inputs, targets)['loss']
        else:
            loss = model.criterion(model(inputs), targets)

        self._synchronize(inputs.device)
        forward_time = time.time() - start

        start = time.time()
        loss.backward()
        self._synchronize(inputs.device)
        backward_time = time.time() - start

        return forward_time, backward_time

    def _prediction_step(self, model, inputs):
        self._synchronize(inputs.device)
        start = time.time()
        if hasattr(model, 'prediction_step'):
            model.prediction_step(inputs)
        elif hasattr(model, 'predictor'):
            model.predictor(model(inputs))
        else:
            model(inputs)
        self._synchronize(inputs.device)
        return time.time() - start



import unittest
from irisml.tasks.create_classification_prompt_generator import Task


class TestCreateClassificationPromptGenerator(unittest.TestCase):
    def test_simple(self):
        outputs = Task(Task.Config()).execute(Task.Inputs())

        results = outputs.generator('random_string')
        self.assertGreater(len(results), 1)
        self.assertTrue(all('random_string' in r for r in results))



"""Clips gradient norm of model parameters."""
import logging
import torch
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, max_norm: float = 1.0):
        self._max_norm = max_norm
        logger.info(f"Enabled gradient clipping. max_norm={max_norm}")

    def on_optimizer_step_start(self, trainer, model, optimizer):
        total_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), self._max_norm)
        if total_norm > self._max_norm:
            logger.debug(f"Clipping the gradients. norm={total_norm}")
        return optimizer



import dataclasses
import logging
import torch
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate accuracy of the given prediction results.

    This task supports only multiclass classification.
    The targets format is Tensor[N] or Tensor[N, 1]
    The prediction results format is torch.Tensor[(N, num_classes)].
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: torch.Tensor
        targets: torch.Tensor

    @dataclasses.dataclass
    class Outputs:
        accuracy: float = 0

    def execute(self, inputs):
        if len(inputs.predictions) == 0:
            raise ValueError("An empty prediction result is provided.")

        if len(inputs.predictions.shape) != 2:
            raise RuntimeError(f"Invalid input shape: {inputs.predictions.shape}")

        _, predicted_max_indexes = inputs.predictions.max(dim=1)
        predicted_max_indexes = predicted_max_indexes.flatten()
        targets = inputs.targets.flatten()

        if targets.shape != predicted_max_indexes.shape:
            raise RuntimeError(f"The predictions or the targets have unexpected shape: {targets.shape} vs {predicted_max_indexes.shape}")

        accuracy = ((predicted_max_indexes == targets).sum() / len(targets)).item()

        logger.info(f"Accuracy: {accuracy:.3f}")
        return self.Outputs(accuracy)



import copy
import random
import unittest
from unittest.mock import MagicMock
import torch
from irisml.tasks.train_with_gradient_cache import Task, GradientCachingTrainer


def fake_transform(x, y):
    return x, y


class TestTrainWithGradientCache(unittest.TestCase):
    def test_split_tensor_or_list(self):
        trainer = GradientCachingTrainer(sub_batch_size=2, model=MagicMock(), lr_scheduler_factory=MagicMock(), optimizer_factory=MagicMock())
        results = trainer._split_tensor_or_list(torch.tensor([[3]] * 4))
        self.assertIsInstance(results, tuple)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].shape, (2, 1))

        results = trainer._split_tensor_or_list((torch.tensor([[3]] * 4), torch.tensor([[2, 2]] * 4)))
        self.assertIsInstance(results, list)
        self.assertEqual(len(results), 2)
        self.assertIsInstance(results[0], tuple)
        self.assertEqual(results[0][0].shape, (2, 1))
        self.assertEqual(results[0][1].shape, (2, 2))
        self.assertEqual(results[1][0].shape, (2, 1))
        self.assertEqual(results[1][1].shape, (2, 2))

        # scalar tensor is returned as-is.
        results = trainer._split_tensor_or_list(torch.tensor(3))
        self.assertEqual(results, torch.tensor(3))

    def test_single_feature(self):
        class Model(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)

            @property
            def criterion(self):
                return torch.nn.CrossEntropyLoss()

            def forward(self, inputs):
                return self._model(inputs).flatten(1)

        self._check_with_model(Model())

    def test_three_features(self):
        class Model(torch.nn.Module):
            class Criterion(torch.nn.Module):
                def forward(self, features, targets):
                    return sum(torch.nn.functional.cross_entropy(f, targets) for f in features)

            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)
                self._model2 = torch.nn.Conv2d(3, 3, 3)
                self._model3 = torch.nn.Conv2d(3, 3, 3)

            @property
            def criterion(self):
                return self.Criterion()

            def forward(self, inputs):
                return self._model(inputs).flatten(1), self._model2(inputs).flatten(1), self._model3(inputs).flatten(1)

        self._check_with_model(Model())

    def test_scalar_feature(self):
        class Model(torch.nn.Module):
            class Criterion(torch.nn.Module):
                def forward(self, features, targets):
                    return torch.sum(features[0]) + torch.sum(features[1])

            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)
                self._param = torch.nn.Parameter(torch.tensor(0.1))

            @property
            def criterion(self):
                return self.Criterion()

            def forward(self, inputs):
                return self._param, self._model(inputs)

        self._check_with_model(Model())

    def test_two_inputs(self):
        class Dataset(torch.utils.data.Dataset):
            def __init__(self, num_examples=8, num_classes=3):
                self._data = [((torch.rand(3, 3, 3), torch.rand(4, 3, 3)), 0)] * num_examples

            def __len__(self):
                return len(self._data)

            def __getitem__(self, index):
                return self._data[index]

        class FakeModel(torch.nn.Module):
            class Criterion(torch.nn.Module):
                def forward(self, features, targets):
                    return sum(torch.sum(f) for f in features)

            def __init__(self):
                super().__init__()
                self._conv = torch.nn.Conv2d(3, 3, 3)
                self._conv2 = torch.nn.Conv2d(4, 3, 3)
                self._received_inputs = []

            def forward(self, inputs):
                self._received_inputs.append(copy.deepcopy(inputs))
                return self._conv(inputs[0]).flatten(1), self._conv2(inputs[1]).flatten(1)

            @property
            def criterion(self):
                return self.Criterion()

        dataset = Dataset()
        model = FakeModel()
        config = Task.Config(batch_size=4, sub_batch_size=1, num_epochs=1)
        outputs = Task(config).execute(Task.Inputs(model, dataset, fake_transform))
        self.assertTrue(all(torch.allclose(dataset[0][0][0], i[0]) for i in outputs.model._received_inputs))
        self.assertTrue(all(torch.allclose(dataset[0][0][1], i[1]) for i in outputs.model._received_inputs))

    def _check_with_model(self, model):
        class Dataset(torch.utils.data.Dataset):
            def __init__(self, num_examples=8, num_classes=3):
                self._data = [(torch.rand(3, 3, 3), random.randrange(num_classes)) for i in range(num_examples)]

            def __len__(self):
                return len(self._data)

            def __getitem__(self, index):
                return self._data[index]

        dataset = Dataset()

        torch.manual_seed(0)
        config = Task.Config(batch_size=4, sub_batch_size=1, num_epochs=1, base_lr=1)
        outputs = Task(config).execute(Task.Inputs(copy.deepcopy(model), dataset, fake_transform))
        model0 = outputs.model

        torch.manual_seed(0)
        config = Task.Config(batch_size=4, sub_batch_size=2, num_epochs=1, base_lr=1)
        outputs = Task(config).execute(Task.Inputs(copy.deepcopy(model), dataset, fake_transform))
        model1 = outputs.model

        torch.manual_seed(0)
        config = Task.Config(batch_size=4, sub_batch_size=4, num_epochs=1, base_lr=1)
        outputs = Task(config).execute(Task.Inputs(copy.deepcopy(model), dataset, fake_transform))
        model2 = outputs.model

        torch.manual_seed(0)
        config = Task.Config(batch_size=4, sub_batch_size=4, num_epochs=3, base_lr=1)
        outputs = Task(config).execute(Task.Inputs(copy.deepcopy(model), dataset, fake_transform))
        model3 = outputs.model

        # All three models are exactly same.
        parameters0 = list(model0.parameters())
        parameters1 = list(model1.parameters())
        parameters2 = list(model2.parameters())
        parameters3 = list(model3.parameters())
        for i in range(len(parameters0)):
            self.assertTrue(torch.allclose(parameters0[i], parameters1[i], atol=0.001))
            self.assertTrue(torch.allclose(parameters0[i], parameters2[i], atol=0.001))
            self.assertFalse(torch.allclose(parameters0[i], parameters3[i], atol=0.001))  # Make sure the allclose() is not too loose.



import collections
import dataclasses
import logging
import statistics
import typing
import sklearn.metrics
import torch
import torchvision.ops
import irisml.core

logger = logging.getLogger(__name__)


class Task(irisml.core.TaskBase):
    """Calculate mean average precision for object detection task results.

    Note that this is a naive implementation. The evaluation results might be different from other libraries.

    Test-time augmentation is not considered.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        predictions: typing.List[torch.Tensor]
        targets: typing.List[torch.Tensor]

    @dataclasses.dataclass
    class Config:
        iou_thresholds: typing.List[float] = dataclasses.field(default_factory=lambda: [0.3, 0.5, 0.75, 0.9])

    @dataclasses.dataclass
    class Outputs:
        results: typing.List[float] = None

    class Evaluator:
        def __init__(self, predictions, targets):
            assert len(predictions) == len(targets)
            self._eval_predictions = collections.defaultdict(lambda: collections.defaultdict(list))
            self._eval_ground_truths = collections.defaultdict(lambda: collections.defaultdict(list))
            for img_index, prediction in enumerate(predictions):
                if len(prediction) > 0 and (len(prediction.shape) != 2 or prediction.shape[1] != 6):
                    logger.warning(f"Invalid prediction shape is detected. Ignoring... {prediction.shape}")
                    continue

                for b in prediction:
                    self._eval_predictions[int(b[0])][img_index].append(b[1:])

            for p in self._eval_predictions.values():
                for q in p.values():
                    q.sort(key=lambda x: -x[1])

            for img_index, target in enumerate(targets):
                if isinstance(target, list):
                    target = torch.tensor(target)

                if target.shape[0] == 0:  # Ignore examples with no bounding box.
                    continue

                if len(target.shape) != 2 or target.shape[1] != 5:
                    logger.warning(f"Invalid target shape is detected. Ignoring... {target.shape}")
                    continue

                for t in target:
                    self._eval_ground_truths[int(t[0])][img_index].append(t[1:])

        def calculate_average_precision(self, iou):
            class_ids = set(list(self._eval_predictions.keys()) + list(self._eval_ground_truths.keys()))
            if not class_ids:
                return 0.0

            average_precisions = []
            for c in class_ids:
                samples = []
                num_ground_truths = sum(len(x) for x in self._eval_ground_truths[c].values())
                for i in set(list(self._eval_predictions[c].keys()) + list(self._eval_ground_truths[c].keys())):
                    samples.extend(self._evaluate_boxes(self._eval_predictions[c][i], self._eval_ground_truths[c][i], iou))

                if num_ground_truths == 0 or not samples or not any(s[0] for s in samples):
                    average_precisions.append(0.0)
                    continue

                recall = float(sum(s[0] for s in samples)) / num_ground_truths
                average_precisions.append(sklearn.metrics.average_precision_score([x[0] for x in samples], [x[1] for x in samples]) * recall)
            return statistics.mean(average_precisions)

        @staticmethod
        def _evaluate_boxes(predictions, ground_truths, iou_threshold):
            """Check if each boxes are correct on a single image.

            Args:
                predictions: [[prob, x1, y1, x2, y2], ...]
                ground_truths: [[x1, y1, x2, y2], ...]
            Returns:
                [(is_correct, probs), ...]
            """
            if not ground_truths:
                return [(False, p[0]) for p in predictions]
            if not predictions:
                return []

            predictions = torch.stack(predictions)
            ground_truths = torch.stack(ground_truths)
            ious = torchvision.ops.box_iou(predictions[:, 1:], ground_truths)
            assert len(ious) == len(predictions)

            # predictions is already sorted by probability.
            results = []
            for i in range(len(ious)):
                best_iou, best_box = ious[i].max(dim=0)  # Find the highest IOU box for each ground truth box.
                if best_iou >= iou_threshold:
                    ious[:, best_box] = -1
                    results.append((True, predictions[i][0]))
                else:
                    results.append((False, predictions[i][0]))
            return results

    def execute(self, inputs):
        evaluator = Task.Evaluator(inputs.predictions, inputs.targets)
        results = [evaluator.calculate_average_precision(iou) for iou in self.config.iou_thresholds]
        logger.info(f"mAP: {results}")
        return self.Outputs(results)

    def dry_run(self, inputs):
        results = [0.0 for _ in self.config.iou_thresholds]
        return self.Outputs(results)



import unittest
import torch
from irisml.tasks.append_classifier import Task


class TestAppendClassifier(unittest.TestCase):
    def test_simple(self):
        model = torch.nn.Sequential(torch.nn.Conv2d(3, 3, 3), torch.nn.Flatten())
        classifier = torch.nn.Linear(147, 5)
        inputs = Task.Inputs(model, classifier)
        outputs = Task(Task.Config('multiclass_classification')).execute(inputs)
        result = outputs.model(torch.rand(1, 3, 9, 9))
        self.assertIsNotNone(result)
        self.assertIsNotNone(outputs.model.criterion)
        self.assertIsNotNone(outputs.model.predictor)



import unittest
from irisml.tasks.build_classification_prompt_dataset import Task


class TestBuildClassificationPromptDataset(unittest.TestCase):
    def test_simple(self):
        class_names = ['class_a', 'class_b', 'class_c']

        def prompt_generator(x):
            return [f'A {x} A', f'B {x} B']

        inputs = Task.Inputs(class_names=class_names, prompt_generator=prompt_generator)
        outputs = Task(Task.Config()).execute(inputs)
        dataset = outputs.dataset

        data = [dataset[i] for i in range(len(dataset))]

        self.assertEqual(set(data), set([('A class_a A', 0), ('A class_b A', 1), ('A class_c A', 2), ('B class_a B', 0), ('B class_b B', 1), ('B class_c B', 2)]))
